from Minter.storage import BasicStorage, MemoryStorage
from Minter.types.wallet import Wallet
from Minter.types.abis.abi import ABI
from Minter.types.nft_data import NFTData
from Minter.transaction_builder import TransactionBuilder
from aiohttp import ClientSession
from web3 import AsyncWeb3, AsyncHTTPProvider
from web3.types import TxReceipt
from typing import List, Union, Callable, Any
from time import perf_counter
import asyncio, json, pathlib

path = pathlib.Path(__file__).parent
NFT_ABI: dict = json.load(open(path/"data/nft_abi.json"))

class Minter:
    def __init__(
        self,
        rpc: str,
        storage: BasicStorage = None,
        semaphore_lock: int = 999999,
    ) -> None:
        self.w3 = AsyncWeb3(AsyncHTTPProvider(rpc))
        self.tx_builder = TransactionBuilder(w3=self.w3)
        self.storage = storage or BasicStorage()
        # for rate limiting purposes
        self.lock = asyncio.Semaphore(semaphore_lock)

    async def connect(self) -> None:
        await self.w3.provider.cache_async_session(ClientSession())
        await self.storage.build()
    
    async def execute(
        self,
        tx_builder_method: Callable,
        *method_args,
        **method_kwargs
    ):
        async with self.lock:
            return await tx_builder_method(
                *method_args,
                **method_kwargs
            )
        
    async def execute_many(
        self,
        tx_builder_method: Callable,
        wallets: List[Wallet] = None,
        bulk: bool = False,
        **method_kwargs
    ) -> List[Any]:
        if not wallets:

            if isinstance(self.storage, MemoryStorage):
                wallets = self.storage.wallets()
            else:
                wallets = await self.storage.wallets()

        _tasks = [
            asyncio.create_task(
                self.execute(tx_builder_method, wallet=wallet, **method_kwargs)
            ) for wallet in wallets
        ]
        result = []

        if bulk:
            result = await asyncio.gather(*_tasks)
        else:
            for task in _tasks:
                result.append(await task)
        
        return result
    
    def process_ids(
        self,
        receipt: TxReceipt,
        rng: int
    ) -> List[int]:
        ids = []

        for log in receipt.get("logs"):
            _ids = [
                self.w3.to_int(hexstr=topic.hex()) for topic in log.get("topics") if 0 < self.w3.to_int(hexstr=topic.hex()) <= rng
            ]
            ids += _ids

        return ids
    
    # raw methods
    async def _save_nft_data(
        self,
        nft_contract: str,
        data: List[Union[None, str]],
        task: bool,
        sleep: int = 5
    ):
        if task:
            return await asyncio.create_task(
                self._save_nft_data(
                    nft_contract=nft_contract,
                    data=data,
                    task=False,
                    sleep=sleep
                )
            )
        
        await asyncio.sleep(sleep)
        supply = await self.execute(
            tx_builder_method=self.tx_builder.get_nft_max_supply,
            contract_address=nft_contract
        )
        receipts: List[TxReceipt] =(await asyncio.gather(
            *[
                asyncio.create_task(self.execute(
                    tx_builder_method=self.tx_builder.get_tx,
                    tx=tx
                )) for tx in data if tx is not None
            ]
        ))
        receipts = [rec for rec in receipts if rec is not None]
        filtered_data = {nft_contract: {}}

        for receipt in receipts:
            receipt_owner = receipt.get("from")

            # success
            if receipt.get("status") == 1:
                nft_ids = self.process_ids(receipt, supply)
                filtered_data[nft_contract][receipt_owner] = nft_ids
            # ??
            elif receipt.get("status") is None:
                continue

            # failed tx
            else:
                continue
        
        # save ids
        if isinstance(self.storage, MemoryStorage):
            self.storage.insert_data(nft_data=NFTData(filtered_data))
        else:
            await self.storage.insert_data(nft_data=NFTData(filtered_data))
    
    # abstract methods
    async def get_balances(
        self,
        wallets: List[Wallet] = None,
        human_readable: bool = False,
        return_raw_data: bool = False,
    ) -> Union[List[float], float]:
        if not wallets:
            if isinstance(self.storage, MemoryStorage):
                wallets = self.storage.wallets()
            else:
                wallets = await self.storage.wallets()
        
        result = await self.execute_many(
            tx_builder_method=self.tx_builder.get_balance,
            bulk=True,
            wallets=wallets,
            human_readable=human_readable,
        )
        
        if return_raw_data:
            return result
        return sum(result)

    async def bulk_fees(
        self,
        fees_wallet: Wallet,
        amount: float,
        hold: int = 3,
        subtract: bool = False,
        wallets: List[Wallet] = None,
        gas: int = 21000,
        tries: int = 3
    ) -> float:
        '''this method is used to distribute fees from fees_wallet to wallets
        Args:
            fees_wallet: Wallet, the wallet that will provide fees
            amount: float, amount of fees every wallet will get IN ETH
            hold: int, sleeps between every fee transfer to keep up with the nonce
            subtract: bool=False,subtract the ETH on every wallet from the amount. aka (keep the wallet at the same amount)
            wallets: List[Wallet]=None, the wallets to distribute fees to, Minter.storage.get_wallets by default
        Returns:
            float, the total amount of ETH sent by the fees_wallet'''
        total = 0
        if not wallets:
            if isinstance(self.storage, MemoryStorage):
                wallets = self.storage.wallets()
            else:
                wallets = await self.storage.wallets()
        
        for wallet in wallets:
            n = amount
            fees_wallet_balance, wallet_balance = await self.get_balances([fees_wallet, wallet], True, True)

            if subtract:
                # e.g amount = 0.001, wallet_balance = 0.002, skips
                n = n - wallet_balance

                if n < 0:
                    continue
            
            if fees_wallet_balance <= n:
                break

            await self.execute(
                tx_builder_method=self.tx_builder.send_eth,
                from_wallet=fees_wallet,
                to=wallet,
                amount=n,
                gas=gas,
                tries=tries
            )
            await asyncio.sleep(hold)
            total += n
        return total
    
    async def mint_bulk(
        self,
        nft_contract: str,
        abi: ABI,
        wallets: List[Wallet] = None,
        args: dict = {},
        price: float = 0,
        gas: int = 250000,
        return_signed_tx: bool = False,
        save_tx_data: bool = True,
        save_in_background: bool = False
    ) -> List[Union[int, asyncio.Task]]:
        start = perf_counter()
        value = self.w3.to_wei(price, "ether")
        result = await self.execute_many(
            tx_builder_method=self.tx_builder.execute_write_function,
            wallets=wallets,
            bulk=True,
            contract_address=nft_contract,
            value=value,
            abi=abi,
            args=args,
            gas=gas,
            return_signed_tx=return_signed_tx
        )
        end = perf_counter()

        if return_signed_tx:
            return result
        
        # saving data and shit
        if save_tx_data:
            # for custom platform mints
            if hasattr(abi, "nft_address"):
                nft_contract = abi.nft_address
            await self._save_nft_data(nft_contract, result, save_in_background)

        return int(end-start)
    
    async def get_nft_balances(
        self,
        nft_contract: str,
        wallets: List[Wallet] = None
    ):
        abi = NFT_ABI
        result = await self.execute_many(
            tx_builder_method=self.tx_builder.execute_read_function,
            wallets=wallets,
            bulk=True,
            contract_address=self.w3.to_checksum_address(nft_contract),
            abi=ABI(
                name="balanceOf",
                abi=abi,
                function_name="balanceOf",
                default_args={"owner": (lambda abi : getattr(abi, "wallet", Wallet(None, None)).address)}
            ),
        )
        
        return sum([r for r in result if isinstance(r, int)])
    
    async def send_nfts(
        self,
        nft_contract: str,
        to: Union[Wallet, str],
        sleep: int = 3,
        limit: int = float("inf"),
        wallets: List[Wallet] = None
    ):
        if isinstance(self.storage, MemoryStorage):
            nft_data = self.storage.nft_data()
        else:
            nft_data = await self.storage.nft_data()
        
        if not wallets:
            if isinstance(self.storage, MemoryStorage):
                wallets = self.storage.wallets()
            else:
                wallets = await self.storage.wallets()
        
        wallets = wallets
        result = []
        n = 0

        for wallet in wallets:
            token_ids = nft_data.get(nft_contract, wallet.address)

            if not token_ids:
                continue

            for _id in token_ids:
                if n >= limit:
                    break

                has_it = await self.execute(
                    tx_builder_method=self.tx_builder.execute_read_function,
                    contract_address=nft_contract,
                    abi=ABI(
                        "ownerOf",
                        abi=NFT_ABI,
                        function_name="ownerOf",
                        default_args={
                            "tokenId": _id
                        }
                    )
                )

                if not has_it:
                    continue

                tx = await self.execute(
                    tx_builder_method=self.tx_builder.execute_write_function,
                    wallet=wallet,
                    contract_address=nft_contract,
                    abi=ABI(
                        "safeTransferFrom",
                        abi=NFT_ABI,
                        function_name="safeTransferFrom",
                        default_args={
                            "from": wallet.address,
                            "to": (to if isinstance(to, str) else to.address),
                            "tokenId": _id
                        }
                    ),
                )

                if tx:
                    result.append(tx)
                    n += 1
                    await asyncio.sleep(sleep)
        return result


from web3 import Account
from Minter.types.wallet import Wallet
from typing import List

def generate_wallets(n: int) -> List[Wallet]:
    res = []

    for _ in range(n):
        account = Account.create()
        res.append(Wallet(
            address=account.address,
            private_key=account._private_key.hex()
        ))
    
    return res


from .types.wallet import Wallet
from .types.abis import (
    ABI,
    OpenseaABI
)
from .types.nft_data import NFTData, ContractData
from .transaction_builder import TransactionBuilder
from .storage import BaseStorage, BasicStorage, MemoryStorage
from .minter import Minter
from .utils import (
    generate_wallets
)


from Minter.types.wallet import Wallet
from Minter.types.abis.abi import ABI
from Minter.types.nft_data import NFTData
from Minter.storage.base_storage import BaseStorage
from typing import Union, List, Dict
import json

class BasicStorage(BaseStorage):
    async def wallets(
        self,
    ) -> Union[None, List[Wallet]]:
        '''returns a list of Minter.types.Wallet'''
        return [Wallet(**wallet) for wallet in (await self.get("wallets"))]
    
    async def abis(
        self,
    ):
        '''returns a list of Minter.types.ABI'''
        return [ABI(**abi) for abi in (await self.get("abis"))]
    
    async def nft_data(
        self,
    ) -> NFTData:
        '''returns all of the nft_data in the db in Minter.types.NFTData object'''
        return NFTData((await self.get("nft_data")))
    
    async def insert_data(
        self,
        data_file: str = None,
        wallets: List[Wallet] = None,
        abis: List[ABI] = None,
        nft_data: NFTData = None
    ) -> Dict:
        '''used to insert any new wallets,abis,nft_data or a json data file
        returns a dict with the updated numbers'''
        data = {
            "wallets": None,
            "abis": None,
            "nft_data": None
        }

        if data_file:
            data = json.load(open(data_file, "r+"))
            wallets = [Wallet(**c) for c in data["wallets"]]
            abis = [ABI(**c) for c in data["abis"]]
            nft_data = NFTData(data["nft_data"])

        if wallets:
            stored_wallets = await self.wallets()
            stored_wallets = stored_wallets + wallets
            await self.set("wallets", [wallet.json for wallet in stored_wallets])
            data["wallets"] = len(stored_wallets)
        
        if abis:
            stored_abis = await self.abis()
            stored_abis = stored_abis + abis
            await self.set("abis", [abi.json for abi in abis])
            data["abis"] = len(stored_abis)

        if nft_data:
            current_data = await self.nft_data()
            current_data.combine(nft_data)
            await self.set("nft_data", current_data.raw_json)
            data["nft_data"] = current_data

        return data
    
    async def export(
        self,
        output_file = "minter_data.json"
    ):
        data = {
            "wallets": await self.get("wallets"),
            "abis": await self.get("abis"),
            "nft_data": await self.get("nft_data")
        }

        with open(output_file, "w+") as f:
            f.write(json.dumps(
                data, indent=4, ensure_ascii=False
            ))
            f.close()
        return output_file


class Wallet:
    def __init__(
        self,
        address: str,
        private_key: str,
    ) -> None:
        '''an abstract Wallet class to store your wallet data
        attrs:
            address: wallet address
            private_key: wallet private key'''

        self.address = address
        self.private_key = private_key
    
    @property
    def json(self):
        return {
            "address": self.address,
            "private_key": self.private_key,
        }
    
    def __str__(self):
        return f'Wallet({self.address}, {self.private_key})'
    
    def __eq__(self, value: object) -> bool:
        return self.address == value if isinstance(value, str) else value.address


from kvsqlite import Client
from typing import Literal, Union, List, Dict

class BaseStorage:
    def __init__(
        self,
        db_name: str = "minter.db",
        table_name: str = "minter",
        *args,
        **kwargs
    ) -> None:
        self.db = Client(db_name, table_name, *args, **kwargs)

    async def build(self):
        keys = ["wallets", "abis", "nft_data"]

        for key in keys:
            if not await self.exists(key):
                await self.set(key, ( {} if key == "nft_data" else []))

    async def rm_rf(self, format_only: bool = True):
        keys = ["wallets", "abis", "nft_data"]

        for key in keys:
            await self.delete(key)        
        if format_only:
            await self.build()

    async def exists(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        return await self.db.exists(key)
    
    async def get(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis.,nft_data not {key}"
        return await self.db.get(key)
    
    async def set(
        self,
        key: Literal["wallets", "abis", "nft_data"],
        data: Union[List[Dict], Dict]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        assert isinstance(data, list) or isinstance(data, dict), "data must be dict or a list of dict's"
        return await self.db.set(key, data)
    
    async def delete(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        return await self.db.delete(key)


from web3 import AsyncWeb3, Account
from web3.contract import AsyncContract
from web3.contract.async_contract import AsyncContractFunction
from Minter.types.wallet import Wallet
from Minter.types.abis.abi import ABI
from typing import Union
import json

class TransactionBuilder:
    
    def __init__(
        self,
        w3: AsyncWeb3
    ) -> None:
        self.w3 = w3

    def get_contract(
        self,
        address: str,
        abi: ABI
    ) -> AsyncContract:
        return self.w3.eth.contract(
            self.w3.to_checksum_address(address),
            abi=abi.abi
        )

    @classmethod
    def sign_tx(
        cls,
        tx,
        wallet: Union[Wallet, str]
    ):
        return Account.sign_transaction(tx, wallet if isinstance(wallet, str) else wallet.private_key)
    
    async def get_tx(
        self,
        tx: str
    ):
        try:
            return await self.w3.eth.get_transaction_receipt(tx)
        except:
            return None
    
    async def execute(
        self,
        tx: dict,
        sign: bool = True,
        signer: Union[Wallet, str] = None,
        broadcast: bool = True,
        tries: int = 3,
    ):
        if sign and signer:
            signed = self.sign_tx(tx, signer)

            if broadcast:
                while tries != 0:
                    try:
                        tx = await self.broadcast(signed)
                        return tx.hex()
                    except:
                        tries -= 1
            return signed
    
    # shortcuts
    async def nonce(
        self,
        address: Union[Wallet, str]
    ):
        return await self.w3.eth.get_transaction_count(address if isinstance(address, str) else address.address)
    
    async def broadcast(
        self,
        tx
    ):
        return await self.w3.eth.send_raw_transaction(tx.raw_transaction)
    
    @property
    async def gas(
        self,
    ):
        return await self.w3.eth.gas_price
    
    @property
    async def _id(
        self,
    ):
        return await self.w3.eth.chain_id

    async def get_balance(
        self,
        wallet: Union[Wallet, str],
        human_readable: bool = False,
    ):
        balance = await self.w3.eth.get_balance(wallet if isinstance(wallet, str) else wallet.address)
        
        if human_readable:
            balance = self.w3.from_wei(balance, "ether")
        return float(balance)
    
    async def send_eth(
        self,
        from_wallet: Wallet,
        to: Union[Wallet, str],
        amount: float,
        return_signed_tx: bool = False,
        gas: int = 21000,
        tries: int = 3,
    ):
        '''used to send ETH (or the main chain token e.g BSCMainToken=BNB) from "from_wallet" to "to"
        args:
            from_wallet: Minter.Wallet
            to: Union[Minter.Wallet, str]
            amount: float, the amount to be send in ETH*
            return_signed_tx: bool = False, weither to return the signed_tx or broadcast it and return tx_hash'''

        amount = self.w3.to_wei(amount, "ether")
        if isinstance(to, str):
            assert self.w3.is_address(to), "'to' argument must be a Minter.Wallet or a vaild evm address"
            to = Wallet(to, "")
        
        gas_price = await self.gas
        data = {
            "from": self.w3.to_checksum_address(from_wallet.address),
            "to": self.w3.to_checksum_address(to.address),
            "value": amount,
            "nonce": (await self.nonce(from_wallet)),
            "gas": gas,
            "maxFeePerGas": gas_price,
            "maxPriorityFeePerGas": gas_price,
            "chainId": await self._id
        }
        return (await self.execute(
            tx=data,
            signer=from_wallet,
            broadcast=not return_signed_tx,
            tries=tries
        ))
    
    async def execute_write_function(
        self,
        wallet: Wallet,
        contract_address: str,
        abi: ABI,
        value: int = 0,
        gas: int = 250000,
        args: dict = {},
        return_signed_tx: bool = False
    ):
        if not abi.function:
            raise ValueError(f"ABI {abi.name} doesn't have a function")
        abi = abi.copy()
        abi.set_wallet(wallet)

        contract = self.get_contract(contract_address, abi)
        args = abi.get_args(args)
        func: AsyncContractFunction = getattr(contract.functions, abi.function["name"])

        tx = await func(**args).build_transaction({
            "from": self.w3.to_checksum_address(wallet.address),
            "value": value,
            "gas": gas,
            "gasPrice": await self.gas,
            "nonce": await self.nonce(wallet)
        })
        result = await self.execute(
            tx,
            sign=True,
            signer=wallet,
            broadcast=not return_signed_tx
        )
        return result
    
    async def execute_read_function(
        self,
        contract_address: str,
        abi: ABI,
        wallet: Wallet = None,
        args: dict = {}
    ):
        if not abi.function:
            raise ValueError(f"ABI {abi.name} doesn't have a function")
        abi = abi.copy()
        abi.set_wallet(wallet=wallet)

        contract = self.get_contract(contract_address, abi)
        func: AsyncContractFunction = getattr(contract.functions, abi.function["name"])
        result = await func(**abi.get_args(args)).call()
        return result
    
    async def get_nft_max_supply(
        self,
        contract_address: str,
    ) -> int:
        try:
            supply = await self.execute_read_function(
                contract_address=self.w3.to_checksum_address(contract_address),
                abi=ABI(
                    "get_max_supply",
                    abi=json.load(open("Minter/data/nft_abi.json", "r")),
                    function_name="maxSupply"
                )
            )
        except:
            # yeah, ik this sucks :'(
            supply = 200_000

        return supply


from Minter.types.abis.abi import ABI
import json, pathlib

path = pathlib.Path(__file__).parent.parent.parent
OPENSEA_ABI = json.load(open(path/"data/opensea_abi.json", "r+"))

class OpenseaABI(ABI):
    def __init__(
        self,
        name,
        nft_address: str,
        quantity: int = 1,
    ):
        self.nft_address = nft_address
        super().__init__(
            name=name,
            abi=OPENSEA_ABI,
            function_name="mintPublic",
            default_args={
                "nftContract": nft_address,
                "feeRecipient": "0x0000a26b00c1F0DF003000390027140000fAa719",
                "minterIfNotPayer": "0x0000000000000000000000000000000000000000",
                "quantity": quantity
            }
        )


from Minter.types.wallet import Wallet
from typing import Union, Dict, Any
import json

class ABI:
    def __init__(
        self,
        name: str,
        abi: Union[Dict, str],
        function_name: str,
        default_args: Dict[str, Any]
    ) -> None:
        if isinstance(abi, str):
            abi = json.loads(abi)

        self.name = name
        self.abi = abi
        self.function: Dict = None
        self.args = default_args

        for func in abi:
            if func.get("name", None) == function_name and func.get("type", None) == "function":
                self.function = func
                break

    def copy(self) -> "ABI":
        return ABI(
            name=self.name,
            abi=self.abi.copy(),
            function_name=self.function.get("name", None),
            default_args=self.args
        )

    def set_wallet(
        self,
        wallet: Wallet
    ):
        self.wallet = wallet
    
    def get_args(
        self,
        data: Dict
    ):
        '''used to get ABI.func args
        
        Args:
            data: dict, any data to overwrite default args'''
        default = self.args
        args = {}

        for arg in default.keys():
            if arg in data.keys():
                target = data[arg]
            else:
                target = default[arg]

            if callable(target):
                target = target(self)

            args[arg] = target

        return args

    def __str__(self):
        return json.dumps(self.json, indent=4, ensure_ascii=False)
    
    @property
    def json(self) -> Dict:
        return {
            "name": self.name,
            "abi": json.dumps(self.abi),
            "function": self.function,
            "args": self.args
        }


from Minter.types import (
    Wallet,
    ABI,
)
from Minter.types.nft_data import NFTData
from typing import Literal, Union, List, Dict

class MemoryStorage:
    def __init__(self):
        self.db = {
            "wallets": [],
            "abis": [],
            "nft_data": NFTData({})
        }
    
    async def build(self) -> None:
        return
    
    def rm_rf(self) -> None:
        self.db["wallets"] = []
        self.db["abis"] = []
        self.db["nft_data"] = {}

    def exists(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        return self.db.exists(key)
    
    def get(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis.,nft_data not {key}"
        return self.db.get(key, None)
    
    def set(
        self,
        key: Literal["wallets", "abis", "nft_data"],
        data: Union[List[Dict], Dict]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        assert isinstance(data, list) or isinstance(data, dict), "data must be dict or a list of dict's"
        self.db[key] = data
    
    def delete(
        self,
        key: Literal["wallets", "abis", "nft_data"]
    ):
        assert key in ["wallets", "abis", "nft_data"], f"key must be one of wallets,abis,nft_data not {key}"
        return self.db.pop(key)
    
    def wallets(self) -> List[Wallet]:
        return self.db.get("wallets", [])
    
    def abis(self) -> List[ABI]:
        return self.db.get("abis", [])
    
    def nft_data(self) -> Union[NFTData, None]:
        return self.db.get("nft_data", None)
    
    def insert_data(
        self,
        wallets: List[Wallet] = None,
        abis: List[ABI] = None,
        nft_data: NFTData = None
    ):
        
        if wallets:
            new = self.wallets() + wallets
            self.set("wallets", new)

        if abis:
            new = self.abis() + abis
            self.set("abis", new)

        if nft_data:
            self.set("nft_data", NFTData)


from setuptools import setup, find_packages

setup(
    name="Minter",
    version="1.0.5",
    description="a lib to mint nft's on any EVM chain",
    long_description=open("readme.md", "r", encoding="utf-8").read(),
    author="SpicyPenguin",
    packages=["Minter", "Minter.types", "Minter.types.abis", "Minter.storage", "Minter.data"],
    include_package_data=True,
    package_data={"": ["*.json"]},
    install_requires=["kvsqlite", "web3"]
)


from dataclasses import dataclass
from typing import Dict, Union, List
import json

@dataclass
class ContractData:
    '''contract data
    
    Args:
        nft_contract: str, the nft contract address
        minters: Dict[str, str], the dict of address-to-txs that minted the nft'''
    nft_contract: str
    minters: Dict[str, List[Union[int]]]

    def get_minter(self, address: str) -> Union[None, List[int]]:
        return self.minters.get(address, None)
    
    def __add__(self, other_contract: "ContractData") -> "ContractData":
        res = ContractData(self.nft_contract, self.minters)
        for address in other_contract.minters:
            if self.minters.get(address, None):
                res.minters[address] = self.minters[address] + other_contract.minters[address]
            else:
                res.minters[address] =  other_contract.minters[address]
        return res

    def __eq__(self, other_contract: Union[str, "ContractData"]) -> bool:
        if isinstance(other_contract, ContractData):
            return self.nft_contract.lower() == other_contract.nft_contract.lower()
        elif isinstance(other_contract, str):
            return self.nft_contract.lower() == other_contract.lower()
        return False

class NFTData:
    def __init__(
        self,
        json_data: Dict[str, Dict[str, List[int]]],
    ):
        self.nfts = {
            nft_contract: ContractData(nft_contract, minters) for nft_contract,minters in json_data.items()
        }

    def get(
        self,
        nft_contract: str,
        address: str
    ) -> Union[None, List[int]]:
        nft_data =ContractData(nft_contract, {})

        for nft in self.nfts:
            if nft_contract.lower() == nft.lower():
                nft_data = self.nfts[nft]
        
        return nft_data.get_minter(address)
    
    def combine(
        self,
        data: Union[Dict, "NFTData"]
    ) -> None:
        if isinstance(data, dict):
            data = NFTData(data)

        for nft_contract in data.nfts:
            # 7BEA
            if self.nfts.get(nft_contract, None):
                # {0x1 = 1}
                cur = self.nfts[nft_contract]
                #print(cur, data.nfts[nft_contract])
                # {0x1 = 1} + {}
                self.nfts[nft_contract] = cur + data.nfts[nft_contract]
            else:
                self.nfts[nft_contract] = data.nfts[nft_contract]
    
    @property
    def json(self):
        return json.dumps({
            nft_address: cd.minters for nft_address, cd in self.nfts.items()
        }, indent=4, ensure_ascii=False)
    
    @property
    def raw_json(self):
        return {nft_address: cd.minters for nft_address, cd in self.nfts.items()}